/* xrpn.c - X-windows RPN calculator */

/*   xrpn.c - Clone of multiplayer game "hunt".
 *   Copyright (C) 2000  Matthew Ogilvie (mmogilvi@flashmail.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* HERE:
 *   Still needed for first "release":
 * 3. inactive color for non-top stack element entries.
 * 4. testing.
 * 5. Think about key bindings.
 * 6. Make sure it doesn't leak even single items of memory/access
 *    freed memory/etc.
 *      Unfortanetly, just doing "gtk_init(...); gtk_exit(0);" creates
 *    some 400 unfreed memory segments.  This app leaves around
 *    800 unfreed, and I'm not sure there is anything I can do
 *    to reduce that.
 */

#include <gtk/gtk.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#ifndef NDEBUG
# define DEBUG_MSGS_ON
#endif /*!NDEBUG*/

#define MAX_STACK_DEPTH 100
#define STR_LEN 1024

#define STATUS_OK 0
#define STATUS_FAIL 1

typedef int Status_t;

/*****************************************/
/* Actual calculator module */

struct CalcData
{ double stack[MAX_STACK_DEPTH];
  int depth;
  double angle2Rad; /* default what to multiply value by to get radians */
  /*int replace;*/
};

#define BUILD_UNARY(funcName,cond,command) \
    Status_t funcName(struct CalcData *calc) \
    { if(calc->depth<1 || (cond)) return STATUS_FAIL; \
      command; \
      return STATUS_OK; \
    }

#define BUILD_BINARY(funcName,cond,command) \
    Status_t funcName(struct CalcData *calc) \
    { if(calc->depth<2 || (cond)) return STATUS_FAIL; \
      command; \
      calc->depth--; \
      return STATUS_OK; \
    }

#define STACK(n) (calc->stack[calc->depth-(n)])

Status_t calcInit(struct CalcData *calc)
{ calc->depth=0;
  calc->angle2Rad=1;
  return STATUS_OK;
}

Status_t calcDestroy(struct CalcData *calc)
{ if(!calc) return STATUS_FAIL;
  return STATUS_OK;
}

Status_t calcInsert(struct CalcData *calc,double number)
{ if(calc->depth>=MAX_STACK_DEPTH)
  { memmove(calc->stack,
            calc->stack+1,
            (MAX_STACK_DEPTH-1)*sizeof(*calc->stack));
    calc->depth--;
  }

  calc->stack[calc->depth++]=number;
  return STATUS_OK;
}

Status_t calcRemove(struct CalcData *calc)
{ if(calc->depth)
  { calc->depth--;
    return STATUS_OK;
  }

  return STATUS_FAIL;
}

double calcTop(struct CalcData *calc)
{ double out;

  if(calc->depth)
  { out=STACK(1);
  }
  else out=0;

  return out;
}

int calcDepth(struct CalcData *calc)
{ return calc->depth;
}

double calcEntry(struct CalcData *calc,int fromTop)
{ if(fromTop>=calc->depth) return 0;
  return calc->stack[calc->depth-1-fromTop];
}

Status_t calcSwap(struct CalcData *calc)
{ double temp;
  if(calc->depth<2) return STATUS_FAIL; \
  temp=STACK(2);
  STACK(2)=STACK(1);
  STACK(1)=temp;
  return STATUS_OK; \
}

BUILD_UNARY(calcSqrt,STACK(1)<0,STACK(1)=sqrt(STACK(1)))
BUILD_UNARY(calcExp,FALSE,STACK(1)=exp(STACK(1)))
BUILD_UNARY(calcChs,FALSE,STACK(1)=-(STACK(1)))
BUILD_UNARY(calcSin,FALSE,STACK(1)=sin(STACK(1)*calc->angle2Rad))
BUILD_UNARY(calcCos,FALSE,STACK(1)=cos(STACK(1)*calc->angle2Rad))
BUILD_UNARY(calcTan,FALSE,STACK(1)=tan(STACK(1)*calc->angle2Rad))
BUILD_UNARY(calcASin,STACK(1)<-1.0 || STACK(1)>1.0,
            STACK(1)=asin(STACK(1))/calc->angle2Rad)
BUILD_UNARY(calcACos,STACK(1)<-1.0 || STACK(1)>1.0,
            STACK(1)=acos(STACK(1))/calc->angle2Rad)
BUILD_UNARY(calcATan,FALSE,STACK(1)=atan(STACK(1))/calc->angle2Rad)
BUILD_UNARY(calc10X,FALSE,STACK(1)=pow(10.0,STACK(1)))
BUILD_UNARY(calc1_X,FALSE,STACK(1)=1.0/STACK(1))
BUILD_UNARY(calcX_2,FALSE,STACK(1)=STACK(1)*STACK(1))
BUILD_UNARY(calcLn,FALSE,STACK(1)=log(STACK(1)));
BUILD_UNARY(calcLog,FALSE,STACK(1)=log10(STACK(1)));

BUILD_BINARY(calcPow,(STACK(2)<0 && STACK(1)!=ceil(STACK(1))),
            STACK(2)=pow(STACK(2),STACK(1)))
BUILD_BINARY(calcAdd,FALSE,STACK(2)+=STACK(1))
BUILD_BINARY(calcSub,FALSE,STACK(2)-=STACK(1))
BUILD_BINARY(calcMul,FALSE,STACK(2)*=STACK(1))
BUILD_BINARY(calcDiv,FALSE,STACK(2)/=STACK(1))

/***************************************************/
/* wrapper around actual calculator (declarations) */

Status_t opQuit(struct CalcData *calc)
{
  gtk_main_quit();

#ifdef DEBUG_MSGS_ON
  printf("Should be quiting\n");
#endif /*DEBUG_MSGS_ON*/
  /*FUTURE: make sure we do good cleanup.*/
  return STATUS_OK;
}

#define SHOW_STACK_COUNT 4

enum operation_e
{ OP_NONE,
  OP_ADD,
  OP_SUB,
  OP_MUL,
  OP_DIV,
  OP_CHS,
  OP_POW,
  OP_EXP,
  OP_SIN,
  OP_COS,
  OP_TAN,
  OP_ASIN,
  OP_ACOS,
  /*OP_ATAN2,*/
  OP_ATAN,
  OP_SQRT,
  OP_POP,
  OP_10X,
  OP_1_X,
  OP_X_2,
  OP_LN,
  OP_LOG,
  OP_SWAP,
  OP_QUIT,

    /* must be manually interpreted */
  OP_DIGIT,
  OP_ENTER
};

struct CharOpTable_s
{ char ch;
  enum operation_e op;
} charOp[]=
 {
   { '+',OP_ADD },
   /*{ '-',OP_SUB },*/
   { 'm',OP_SUB },
   { '*',OP_MUL },
   { '/',OP_DIV },
   /*{ 'c',OP_CHS },*/
   { 'h',OP_CHS },
   { 'p',OP_POW },
   { 'o',OP_POP },
   { 'x',OP_EXP },
   { 's',OP_SIN },
   { 'c',OP_COS },
   { 't',OP_TAN },
   { 'S',OP_ASIN },
   { 'C',OP_ACOS },
   { 'T',OP_ATAN },
   { 'r',OP_SQRT },
   { 'q',OP_QUIT },
   { 'l',OP_LN },
   /*{ '\0',OP_ATAN2 },*/
   { '\0',OP_NONE }
  };

struct OpTable_s
{ enum operation_e op;
  Status_t (*func)(struct CalcData *calc);
} opTable[]=
 {
   { OP_ADD,calcAdd },
   { OP_SUB,calcSub },
   { OP_MUL,calcMul },
   { OP_DIV,calcDiv },
   { OP_CHS,calcChs },
   { OP_POW,calcPow },
   { OP_EXP,calcExp },
   { OP_SIN,calcSin },
   { OP_COS,calcCos },
   { OP_TAN,calcTan },
   { OP_ASIN,calcASin },
   { OP_ACOS,calcACos },
   { OP_ATAN,calcATan },
   /*{ OP_ATAN2,calcATan2 },*/
   { OP_SQRT,calcSqrt },
   { OP_POP,calcRemove },
   { OP_10X,calc10X },
   { OP_1_X,calc1_X },
   { OP_X_2,calcX_2 },
   { OP_LN,calcLn },
   { OP_LOG,calcLog },
   { OP_SWAP,calcSwap },
   { OP_QUIT,opQuit },
   { OP_NONE }
 };


struct button_s
{ char *label;
  enum operation_e op;
  int xsz,ysz;
} buttonTable[]=
 {
   { "sin", OP_SIN,1,1 },
   { "cos", OP_COS,1,1 },
   { "tan", OP_TAN,1,1 },
   { "", OP_NONE,1,1 },
   { "pop", OP_POP,1,1 }, /* New line */
   { "asin", OP_ASIN,1,1 },
   { "acos", OP_ACOS,1,1 },
   { "atan", OP_ATAN,1,1 },
   { "", OP_NONE,1,1 },
   { "quit", OP_QUIT,1,1 }, /* new line */
   { "sqrt", OP_SQRT,1,1 },
   { "exp", OP_EXP,1,1 },
   { "10^x", OP_10X,1,1 },
   { "y^x", OP_POW,1,1 },
   { "1/x", OP_1_X,1,1 }, /* new line */
   { "x^2", OP_X_2,1,1 },
   { "ln", OP_LN,1,1 },
   { "log", OP_LOG,1,1 },
   { "", OP_NONE,1,1 },
   { "x<>y", OP_SWAP,1,1 }, /* new line */
   { "chs", OP_CHS,1,1 },
   { "7", OP_DIGIT,1,1 },
   { "8", OP_DIGIT,1,1 },
   { "9", OP_DIGIT,1,1 },
   { "/", OP_DIV,1,1 }, /* new line */
   { "", OP_ENTER,1,3 },
   { "4", OP_DIGIT,1,1 },
   { "5", OP_DIGIT,1,1 },
   { "6", OP_DIGIT,1,1 },
   { "*", OP_MUL,1,1 }, /* new line */
   { "", OP_NONE,1,1 },
   { "1", OP_DIGIT,1,1 },
   { "2", OP_DIGIT,1,1 },
   { "3", OP_DIGIT,1,1 },
   { "-", OP_SUB,1,1 }, /* new line */
   { "", OP_NONE,1,1 },
   { "0", OP_DIGIT,1,1 },
   { ".", OP_DIGIT,1,1 },
   { "e", OP_DIGIT,1,1 },
   { "+", OP_ADD,1,1 }, /* new line (sort of) */
   { NULL, OP_NONE }
 };


/**********************/
/* some dialog stuff */
struct CalcDialog_s
{ /* int autoInsert; / * if 1, then on next operation, first push new entry */
  int newOne;
  int updating;
  struct CalcData *calculator;
  GtkWidget *entry[SHOW_STACK_COUNT];
  /*GtkWidget *extras[SHOW_STACK_COUNT];*/
};

void updateText(struct CalcDialog_s *calcDialog)
   /* newOne==1 iff the we are creating a new element not yet
    * put onto the stack. */
{ char buf[1024];
  int a;
  int count;

#ifdef DEBUG_MSGS_ON
  printf("updating entry to %g\n",calcEntry(calcDialog->calculator,0));
#endif /*DEBUG_MSGS_ON*/
  assert(calcDialog->newOne==1 || calcDialog->newOne==0);

  calcDialog->updating=TRUE;
  count=calcDepth(calcDialog->calculator);
  for(a=0;a<count && a+calcDialog->newOne<SHOW_STACK_COUNT;a++)
  { GtkWidget *entry=calcDialog->entry[a+calcDialog->newOne];
    int changed;
    sprintf(buf,"%.15g",calcEntry(calcDialog->calculator,a));
    changed=strcmp(buf,gtk_entry_get_text(GTK_ENTRY(entry)));
    if(changed) gtk_entry_set_text(GTK_ENTRY(entry),buf);
    if(a+calcDialog->newOne==0)
    { /*OBS calcDialog->needRegionSelect=changed; */
      /*OBS if(!changed) ;*/
      gtk_entry_select_region(GTK_ENTRY(entry),0,strlen(buf));
#ifdef DEBUG_MSGS_ON
      printf("selecting entry to %g\n",calcEntry(calcDialog->calculator,0));
#endif /*DEBUG_MSGS_ON*/
    }
  }
  a+=calcDialog->newOne;
  for(;a<SHOW_STACK_COUNT;a++)
  { gtk_entry_set_text(GTK_ENTRY(calcDialog->entry[a]),"");
  }
  calcDialog->updating=FALSE;
}

/**********************/
/* implementation of calculator wrapper */

enum operation_e lookupCharOp(char ch)
{ int a;
  for(a=0;charOp[a].op!=OP_NONE;a++)
  { if(ch==charOp[a].ch) return charOp[a].op;
  }
  return OP_NONE;
}

char revLookupCharOp(enum operation_e op)
{ int a;
  for(a=0;charOp[a].op!=OP_NONE;a++)
  { if(op==charOp[a].op) return charOp[a].ch;
  }
  return '\0';
}

/**********************/
struct OpTable_s *lookupOpTbl(enum operation_e op)
{ int a;
  for(a=0;opTable[a].op!=OP_NONE;a++)
  { if(opTable[a].op==op) return opTable+a;
  }
  return NULL;
}

struct OpTable_s *lookupOpTblChar(char ch)
{ enum operation_e op;
  op=lookupCharOp(ch);
  if(op==OP_NONE) return NULL;
  return lookupOpTbl(op);
}

Status_t invokeOp(struct OpTable_s *opEntry,struct CalcDialog_s *calcDialog)
{ Status_t stat;
  if(calcDialog->newOne)
  { const char *strVal=gtk_entry_get_text(GTK_ENTRY(calcDialog->entry[0]));
    double val=atof(strVal);
    calcInsert(calcDialog->calculator,val);
    calcDialog->newOne=0;
  }
  stat=(*opEntry->func)(calcDialog->calculator);
  if(stat==STATUS_OK)
  { calcDialog->newOne=0;
    updateText(calcDialog);
  }
#ifdef DEBUG_MSGS_ON
  else
  { printf("Operation failed\n");
  }
#endif /*DEBUG_MSGS_ON*/
  return stat;
}

/**********************/
/* the rest of the dialog management functions */

void entry_activate(GtkWidget *widget,gpointer data)
{ gchar *entry_text;
  double value;
  struct CalcDialog_s *calcDialog=(struct CalcDialog_s *)data;

  entry_text=gtk_entry_get_text(GTK_ENTRY(calcDialog->entry[0]));
  /*FUTURE: make sure it is a double */
  value=atof(entry_text);
#ifdef DEBUG_MSGS_ON
  printf("Inserting %.15g\n",value);
#endif /*DEBUG_MSGS_ON*/

  calcInsert(calcDialog->calculator,value);
  calcDialog->newOne=0;

  updateText(calcDialog);
}

gint key_press_event(GtkWidget *widget, GdkEventKey *event,gpointer data)
{ const char *keystring=event->string;
  struct OpTable_s *opEntry;
  struct CalcDialog_s *calcDialog=(struct CalcDialog_s *)data;

#ifdef DEBUG_MSGS_ON
  printf("pressed \"%s\"\n",event->string);
#endif /*DEBUG_MSGS_ON*/

  if(keystring[0]=='\n' || keystring[0]=='\r')
  { entry_activate(widget,calcDialog);
    gtk_signal_emit_stop_by_name(GTK_OBJECT(widget),"key_press_event");
#ifdef DEBUG_MSGS_ON
    printf("Manual activate\n");
#endif /*DEBUG_MSGS_ON*/
    return TRUE;
  }
  else if(keystring[0])
  { opEntry=lookupOpTblChar(keystring[0]);
    if(opEntry!=NULL && opEntry->func!=NULL)
    { Status_t stat;
#ifdef DEBUG_MSGS_ON
      printf("Invoking function\n");
#endif /*DEBUG_MSGS_ON*/
      stat=invokeOp(opEntry,calcDialog);
      if(stat==STATUS_OK)
      { gtk_signal_emit_stop_by_name(GTK_OBJECT(widget),"key_press_event");
#ifdef DEBUG_MSGS_ON
        printf("Successfull invocation\n");
#endif /*DEBUG_MSGS_ON*/
        return TRUE;
      }
    }
  }

  return FALSE;
}

void entry_changed(GtkWidget *widget,gpointer data)
{ gchar *entry_text;
  struct CalcDialog_s *calcDialog=(struct CalcDialog_s *)data;

  entry_text=gtk_entry_get_text(GTK_ENTRY(calcDialog->entry[0]));
    /*FUTURE: some simplistic verification
     * (at least don't allow illegal chars in; possibly implement
     *  some kind of "could this be a prefix of some string that
     *  could match this RE?")
     */
#ifdef DEBUG_MSGS_ON
  printf("entry_changed() updating %d newOne %d depth %d\n",
         calcDialog->updating,
         calcDialog->newOne,
         calcDepth(calcDialog->calculator));
#endif /*DEBUG_MSGS_ON*/
  if(!calcDialog->updating && calcDialog->newOne==0)
  { calcDialog->newOne=1;
    updateText(calcDialog);
  }
}

void handle_button(GtkWidget *buttonWidget,gpointer calcDialogVPtr)
{ struct CalcDialog_s *calcDialog=(struct CalcDialog_s *)calcDialogVPtr;
  gpointer buttonVptr;
  struct button_s *button;
  struct OpTable_s *opEntry;
  int position;

  assert(button!=NULL);

  buttonVptr=gtk_object_get_data(GTK_OBJECT(buttonWidget),"button_s handler");
  assert(buttonVptr!=NULL);
  if(buttonVptr==NULL) return;

  button=(struct button_s *)buttonVptr;

  switch(button->op)
  { case OP_ENTER:
         /*NOTE: would it make more sense to emit some
          * kind of signal or fake event? */
      entry_activate(calcDialog->entry[0],calcDialog);
    break;
    case OP_DIGIT:
          /*NOTE: this really feels ugly.  Is it even legitamate to
           * mess with GTK_ENTRY vie gtk_editable*()?  */
      gtk_editable_delete_selection(GTK_EDITABLE(calcDialog->entry[0]));
      position=gtk_editable_get_position(GTK_EDITABLE(calcDialog->entry[0]));
      gtk_editable_insert_text(GTK_EDITABLE(calcDialog->entry[0]),
                               button->label,1,&position);
      gtk_editable_set_position(GTK_EDITABLE(calcDialog->entry[0]),position);
      entry_changed(calcDialog->entry[0],calcDialog);
    break;
    case OP_NONE:
      assert(1==0);
    break;
    default:
      opEntry=lookupOpTbl(button->op);
      if(opEntry==NULL || opEntry->func==NULL)
      {
#ifdef DEBUG_MSGS_ON
        printf("Unknown operation %d\n",button->op);
#endif /*DEBUG_MSGS_ON*/
        return;
      }
      invokeOp(opEntry,calcDialog);
  }
  gtk_widget_grab_focus(calcDialog->entry[0]);
/*FUTURE:consider ways of getting desired "replace-entry" semantics
 *   without trashing clipboard. */
}

void packTable(struct CalcDialog_s *calcDialog,
              GtkWidget *table,int width,int height,struct button_s *buttons)
{ int x,y;
  int count;
  GtkWidget *box;
  GtkWidget *button;
  const char *cPtr;
  char tempLabel[STR_LEN+1];
  char ch;
  int inpos,outpos;

  gtk_table_set_row_spacing(GTK_TABLE(table),2,2);
  count=0;
  for(y=0;y<height;y++)
  { for(x=0;x<width;x++,count++)
    { if(!buttons[count].label) continue;
      switch(buttons[count].op)
      { case OP_NONE:
          /*nothing*/
        break;
        case OP_ENTER:
          box=gtk_vbox_new(TRUE,0);
          for(cPtr="ENTER";*cPtr;cPtr++)
          { char s[2];
            GtkWidget *label;
            s[1]='\0';
            s[0]=*cPtr;
            label=gtk_label_new(s);
            gtk_box_pack_start(GTK_BOX(box),label,FALSE,FALSE,0);
            gtk_widget_show(label);
          }
          button=gtk_button_new();
          gtk_signal_connect(GTK_OBJECT(button), "clicked",
                GTK_SIGNAL_FUNC(handle_button), calcDialog);
          gtk_object_set_data(GTK_OBJECT(button), "button_s handler",
                &buttons[count]);
          gtk_container_add(GTK_CONTAINER(button),box);
          gtk_widget_show(box);
          gtk_table_attach(GTK_TABLE(table),button,
                x,x+buttons[count].xsz,y,y+buttons[count].ysz,
                GTK_EXPAND|GTK_FILL,GTK_EXPAND|GTK_FILL,1,1);
          gtk_widget_show(button);
        break;
        default:
          if(buttons[count].label[0]=='\0')
          { fprintf(stderr,"Internal inconsistency\n");
           continue;
          }
          ch=revLookupCharOp(buttons[count].op);
          sprintf(tempLabel,"%.*s",STR_LEN,buttons[count].label);
          for(inpos=outpos=0;
              buttons[count].label[inpos]&&outpos<STR_LEN;
              inpos++)
          { if(buttons[count].label[inpos]==ch && inpos && outpos<STR_LEN-3)
            { sprintf(tempLabel+outpos,"(%c)",ch);
              outpos+=3;
            }
            else
            { tempLabel[outpos++]=buttons[count].label[inpos];
            }
          }
          tempLabel[outpos]='\0';
          button=gtk_button_new_with_label(tempLabel);
          gtk_signal_connect(GTK_OBJECT(button), "clicked",
                GTK_SIGNAL_FUNC(handle_button), calcDialog);
          gtk_object_set_data(GTK_OBJECT(button), "button_s handler",
                &buttons[count]);
          gtk_table_attach(GTK_TABLE(table),button,
                x,x+buttons[count].xsz,y,y+buttons[count].ysz,
                GTK_EXPAND|GTK_FILL,GTK_EXPAND|GTK_FILL,1,1);
          gtk_widget_show(button);
      }
    }
  }
}

/**********************/

int main(int argc, char *argv[])
{
  GtkWidget *window;
  GtkWidget *box;
  GtkWidget *stackBox;
  GtkWidget *entry;
  GtkWidget *table;
  GtkWidget *sep;
  struct CalcData calculator;
  struct CalcDialog_s calcDialog;
  int a;

#ifndef NDEBUG
  mtrace();
#endif

    /* Init calculator */
  calcInit(&calculator);
  calcDialog.calculator=&calculator;
  calcDialog.newOne=0;
  calcDialog.updating=FALSE;

    /* Init dialog */
  gtk_init(&argc, &argv);

  window=gtk_window_new(GTK_WINDOW_TOPLEVEL);

  gtk_signal_connect(GTK_OBJECT(window),"delete_event",
                     GTK_SIGNAL_FUNC(gtk_main_quit),NULL);
  gtk_container_set_border_width(GTK_CONTAINER(window),5);
  box=gtk_vbox_new(FALSE,5);

    /* stack display */
  stackBox=gtk_vbox_new(FALSE,0);
  entry=gtk_entry_new();
  calcDialog.entry[0]=entry;
#if 0
  gtk_signal_connect(GTK_OBJECT(entry),
		"activate",
		GTK_SIGNAL_FUNC(entry_activate),
		&calcDialog);
#endif /*0*/
  gtk_signal_connect(GTK_OBJECT(entry),
                "key_press_event",
                GTK_SIGNAL_FUNC(key_press_event),
                &calcDialog);
  gtk_signal_connect(GTK_OBJECT(entry),
                "changed",
                GTK_SIGNAL_FUNC(entry_changed),
                &calcDialog);

  gtk_widget_show(GTK_WIDGET(entry));
  gtk_box_pack_end(GTK_BOX(stackBox),entry,FALSE,FALSE,0);
  for(a=1;a<SHOW_STACK_COUNT;a++)
  { entry=gtk_entry_new();
    calcDialog.entry[a]=entry;
    gtk_widget_show(entry);
    gtk_entry_set_editable(GTK_ENTRY(entry),FALSE);
    gtk_box_pack_end(GTK_BOX(stackBox),entry,FALSE,FALSE,0);
  }
  gtk_widget_show(GTK_WIDGET(stackBox));
  gtk_box_pack_start(GTK_BOX(box),stackBox,TRUE,TRUE,5);

     /* FUTURE: buttons */
  sep=gtk_hseparator_new();
  gtk_widget_show(GTK_WIDGET(sep));
  gtk_box_pack_start(GTK_BOX(box),sep,TRUE,TRUE,5);

  table=gtk_table_new(8,5,TRUE);
  packTable(&calcDialog,table,5,8,buttonTable);
  gtk_box_pack_start(GTK_BOX(box),table,TRUE,TRUE,5);
  gtk_widget_show(table);

  gtk_widget_show(GTK_WIDGET(box));

  gtk_container_add(GTK_CONTAINER(window),box);
  gtk_widget_show(window);

  gtk_widget_grab_focus(calcDialog.entry[0]);

  gtk_main();

  gtk_widget_destroy(GTK_WIDGET(window));
  /*FUTURE: how to destroy widgets.*/

  calcDestroy(&calculator);

  //mallocPrintStats();

  gtk_exit(0);

  return 0;
}
